%
% HEOM.m
%
% This function calculates the population of the average eigenfunctions
% using the HEOM method following the implementation of Ishizaki of
% J. Chem. Phys.  130  234111  (2009)
function LL=HEOM
    term=5; % Depth of hierachy
    term=term+1;
    icm2ifs=2.99792458e-5; % cm/fs
    deltat=10; % Timestep in fs
    factor=icm2ifs*2*pi*deltat;
    w1=70;w2=0; % Frequencies in cm-1
    J=36;            % Coupling in cm-1
    T=150; % Kelvin
    kBT=T*0.6950389; % cm-1
    beta=1/kBT;
    tau=50; % Correlation time in fs
    g1=1/tau/factor*deltat;g2=g1; % Correlation times in cm-1
    sigma=25; % Magnitude of fluctuations in cm-1
    l1=sigma^2/kBT/2;l2=sigma^2/kBT/2;     % Reorganization energy cm-1
    time=4000; % Length of trajectory in fs
    
    % Site boltzman
    bzman=exp(-(w1-w2)*beta);
    bzman1=bzman/(1+bzman);
    bzman2=1/(1+bzman);
    % Quantum boltzman
    [ ve, ev]=eig([w1 J;J w2]);
    bz=exp(-(ev(1,1)-ev(2,2))*beta);
    bz1=bz/(1+bz);
    bz2=1/(1+bz);
    pop11=ve(1,1)^2/(ve(1,1)^2+ve(1,2)^2);
    pop12=ve(2,1)^2/(ve(2,2)^2+ve(2,1)^2);
    pop22=ve(2,2)^2/(ve(2,2)^2+ve(2,1)^2);
    pop21=ve(1,2)^2/(ve(1,1)^2+ve(1,2)^2);
    pop1=pop11*bz1+pop12*bz2;
    pop2=pop22*bz2+pop21*bz1;
    [ pop11 pop12 pop1 bz1 bz2 bzman1 bzman2 ];
    
    elements=(term+1)*term/2
    % Test indexing scheme
    for k=0:term-1;
        for l=0:term-1;
            if (k+l)<=term-1
                index=k*(term)+l-k*(k-1)/2+1;
                myindex(k,l,term);
            end
        end
    end

    NN=elements;
    rho=zeros(NN*4,1);
    rho(1)=1;
    rho2=zeros(NN*4,1);
    rho2(2)=1;

    % Find avearge eigenstates
    H=[ w1 J ; J w2 ];
    [v,E]=eig(H);
    rho3=zeros(NN*4,1);
    rho4=zeros(NN*4,1);
    rho3(1:4)=[ v(1,1)*v(1,1) ; v(2,1)*v(2,1); v(1,1)*v(2,1); v(2,1)*v(1,1) ];
    rho4(1:4)=[ v(1,2)*v(1,2) ; v(2,2)*v(2,2); v(1,2)*v(2,2); v(2,2)*v(1,2) ];
    exppop=exp(-(E(2,2)-E(1,1))/kBT)/2;
    
    
    % Build basis matrices
    L = [ 0 0 -J J ; 0 0 J -J ; -J J w1-w2 0 ; J -J 0 w2-w1 ];
    phi1=[ 0 0 0 0 ; 0 0 0 0 ; 0 0 1 0 ; 0 0 0 -1 ];
    phi2=[ 0 0 0 0 ; 0 0 0 0 ; 0 0 -1 0 ; 0 0 0 1 ];
    theta1=2*l1/beta*phi1-[2 0 0 0 ; 0 0 0 0; 0 0 1 0; 0 0 0 1]*l1*g1*i;
    theta2=2*l2/beta*phi2-[0 0 0 0 ; 0 2 0 0; 0 0 1 0; 0 0 0 1]*l2*g2*i;
    unit= [ 1 0 0 0 ; 0 1 0 0 ; 0 0 1 0 ; 0 0 0 1 ];
    
    % Build big matrix
    % Diagonal
    for k=0:term
        for l=0:term
            if (k+l+1)<term
                ind=myindex(k,l,term)-1;
                LL(4*ind+1:4*ind+4,4*ind+1:4*ind+4)=-i*L-k*g1*unit-l*g2*unit;
            end
            % Terminator
            if (k+l+1)==term
                ind=myindex(k,l,term)-1;
                LL(4*ind+1:4*ind+4,4*ind+1:4*ind+4)=-i*L-k*g1*unit-l*g2*unit; % Terminator
            end
        end
    end
    % Phi and theta couplings
    for k=0:term
        for l=0:term
            if (k+l+2)<=term
                ind=myindex(k,l,term)-1;
                if l<term-1
                    index=myindex(k,l+1,term)-1;
                    LL(4*ind+1:4*ind+4,4*index+1:4*index+4)=i*phi2;
                    LL(4*index+1:4*index+4,4*ind+1:4*ind+4)=(l+1)*i*theta2;
                end
                if k<term-1
                    index=myindex(k+1,l,term)-1;
                    LL(4*ind+1:4*ind+4,4*index+1:4*index+4)=i*phi1;
                    LL(4*index+1:4*index+4,4*ind+1:4*ind+4)=(k+1)*i*theta1;
                end
            end
        end
    end

    % Solve for a single timestep
    u=expm(LL*factor);
    rho0=rho;
    rhoHE=rho4;
    for t=0:time/deltat
        x(t+1)=t*deltat;
        r1(t+1)=rho'*rho0;
        r2(t+1)=rho2'*rho0;
        r3(t+1)=pop1; %bzman1;
        r4(t+1)=pop2;
        r5(t+1)=rho3'*rhoHE;
        r6(t+1)=rho4'*rhoHE;
        rho0=u*rho0;
        rhoHE=u*rhoHE;
        
    end
    % Plot the population of the average eigenstate
    figure(1)
    plot(x,real(r6),'r','LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
    axis([ 0 time 0 1])
end

% Subroutine for indexing of hierachy
function ind=myindex(a,b,N)
    ind=a*(N)+b-a*(a-1)/2+1;
end