%
% ImprovedScaled.m
%
% This matlab function calculates the population of the
% instantaneous eigenfunctions using the Kleinekathoefer
% Improved Scaling Method (J. Chem. Phys.  136  214101  (2012))
function ImprovedScaled()
tic
N=2; % Number of sites (only fully tested for 2)
r=10000; % Number of realizations
dt=10; % Time step in fs

icm2ifs=2.99792458e-5; % cm/fs
deltat=dt; % Timestep in fs
factor=icm2ifs*2*pi*deltat;
T=150; % Temperature in Kelvin 
kBT=T*0.6950389; % Temperature in cm-1
time=4000; % Length in fs

dw0=70; % Energy Gap (cm-1)
sigma=25; % Fluctuation Magnitude (cm-1)
J=36; % Nearest Neighbor Coupling (cm-1)
tau=50; % Time scale (fs)

% Initialize variables
a=exp(-dt/tau);
b=sqrt(1-a*a);
H0=diag(J*ones(N-1),1)+diag(J*ones(N-1),-1);
H0(1,1)=dw0;
w=randn(N,1)*sigma;
PU=zeros(size(0:dt:time));
PSLOC=zeros(size(0:dt:time));
% Loop over realizations
for rz=1:r
    % Initialize frequencies
    w=randn(N,1)*sigma;
    % Initialize real wavefunction
    phi=zeros(N,1);
    H=H0+diag(w);
    [C,E]=eig(H);
    [C0,E0]=eig(H);
    phi=C(:,N); % Excitation
    Cold=C;
    Eold=E;
    phiLOC=[1 ; 0];
    state=N;
    phiB=phi;
    phiU=phi;
    PU(1)=PU(1)+1;
    PB(1)=PB(1)+1;
    PULOC(1)=PULOC(1)+1;
    step=2;
    % Transfer to instantaneous eigenbasis
    phiU=Cold'*phiU;
    phiB=Cold'*phiB;
    phiULOC=Cold'*phiLOC;
    % Loop over times
    for t=dt:dt:time
        % Generate new hamiltonian
        w=w*a+b*randn(N,1)*sigma;
        H=H0+diag(w);
        [C,E]=eig(H);
        % Find Js
        [J,JB,JU]=find_J(H-H0,C,E,kBT,N,phiB,phiU);
        % Update wavefunctions in eigenbasis
        phiB=updateWave(phiB,E,JB,N,dt);
        phiU=updateWave(phiU,E,JU,N,dt);  
	% For sites
        [J,JB,JU]=find_J(H-H0,C,E,kBT,N,phiB,phiULOC);
        % Update wavefunctions in eigenbasis
        phiULOC=updateWave(phiULOC,E,JU,N,dt);
        % Find Populations of the originally populated eigenstate
        PU(step)=PU(step)+abs(C(:,N)'*C*phiU)^2;
        PB(step)=PB(step)+abs(C(:,N)'*C*phiB)^2;
	% Find population in site basis
        PULOC(step)=PULOC(step)+abs([1 0]*C*phiULOC)^2;
        step=step+1;
    end
end
PU=PU/r;
PB=PB/r
PULOC=PULOC/r;
% Make Figure of Kleinekathoefer Improved Scaling
figure(1)
plot(0:dt:time,PU,'m','LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
axis([ 0 time 0 1])
toc
end

% Sub routine for scaling non-adiabatic coupling
function [J,JB,JU]=find_J(H,C,E,kBT,N,phiB,phiU)
J=zeros(N,N);
% Find normal J
for a=1:N
    for b=1:N
        for j=1:N
            J(a,b)=J(a,b)+C(a,j)*C(b,j)*H(j,j);
        end
        DE=-(E(a,a)-E(b,b));
        Jt(a,b)=J(a,b)*sqrt(2/(1+exp(-DE/kBT)));
    end
end
% Find corrected JB (Bastida)
for a=1:N
    for b=1:N
        if (E(a,a)>E(b,b))
            ca=abs(phiB(a));
            cb=abs(phiB(b));
            JB(a,b)=cb*Jt(a,b)-ca*Jt(b,a);
            JB(b,a)=cb*Jt(a,b)-ca*Jt(b,a);
        end
        JB(a,a)=J(a,a);
    end
end
% Find corrected JU (Kleinekathoefer)
for a=1:N
    for b=1:N
        if (E(a,a)>E(b,b))
            ca=abs(phiU(a));
            cb=abs(phiU(b));
            if ca~=cb
                JU(a,b)=(cb*Jt(a,b)-ca*Jt(b,a))/(cb-ca);
                JU(b,a)=(cb*Jt(a,b)-ca*Jt(b,a))/(cb-ca);
            else
                JU(a,b)=(Jt(a,b)+Jt(b,a))/2;
                JU(b,a)=(Jt(a,b)+Jt(b,a))/2;
            end
        end
        JU(a,a)=J(a,a);
    end
end
           
end

% Subroutine for updating wavefunctions
function phinew=updateWave(phi,E,J,N,dt)
icm2ifs=2.99792458e-5; % cm/fs
deltat=dt; % Timestep in fs
factor=icm2ifs*2*pi*deltat;

phinew=expm(-i*factor*(E+J))*phi;
end
