%
% SHnew.m
%
% This matlab function calculates the population of the
% instantaneous eigenfunctions using the SESH method 
% described in this publication. The NISE implementations of
% J. Phys. Chem. B  110  22910-22916  (2006) is also included,
% which corresponds to Ehrenfest without feedback.
function SHnew()
tic
N=2; % Number of sites (only fully tested for 2)
r=10000; % Number of realizations
dt=10.0; % Time step in fs

icm2ifs=2.99792458e-5; % cm/fs
deltat=dt; % Timestep in fs
factor=icm2ifs*2*pi*deltat;
T=150; % Kelvin 
kBT=T*0.6950389; % cm-1
time=4000; % Length in fs

dw0=70; % Energy Gap (cm-1)
sigma=25; % Fluctuation Magnitude
J=36; % Nearest Neighbor Coupling
tau=50; % Time scale in fs

% Initialize variables
a=exp(-dt/tau)
b=sqrt(1-a*a);
H0=diag(J*ones(N-1),1)+diag(J*ones(N-1),-1);
H0(1,1)=dw0;
w=randn(N,1)*sigma;
P=zeros(size(0:dt:time));
PSH=zeros(size(0:dt:time));
PLOC=zeros(size(0:dt:time));
% Loop over realizations
for rz=1:r
    % Initialize frequencies
    w=randn(N,1)*sigma;
    % Initialize real wavefunction
    phi=zeros(N,1);
    H=H0+diag(w);
    [C,E]=eig(H);
    phi=C(:,N); % Excitation
    phiLOC=[1 ; 0];
    state=N;
    phiA=phi;
    P(1)=P(1)+1;
    PLOC(1)=PLOC(1)+1;
    PSH(1)=PSH(1)+1;
    step=2;
    % Loop over times
    for t=dt:dt:time
        % Generate new hamiltonian
        w=w*a+b*randn(N,1)*sigma;
        H=H0+diag(w);
        % Do normal NISE propagation and propagation of primary
        % wavefunction
        phin=expm(-i*H*factor)*phi;
        phiLOC=expm(-i*H*factor)*phiLOC;
        [C,E]=eig(H);
        % Check for surface hopping
        oldstate=state;
        state=find_present_state(oldstate,C,phiA);
        state=hopping(state,oldstate,C,phiA,phin,E,kBT,dt);
        phiA=C(:,state);
        P(step)=P(step)+abs(C(:,N)'*phin)^2;
        PLOC(step)=PLOC(step)+abs([1 0]*phiLOC)^2;
        PSH(step)=PSH(step)+abs(C(:,N)'*phiA)^2;
        step=step+1;
        phi=phin;
    end
end
P=P/r;
PSH=PSH/r;
PLOC=PLOC/r;
figure(1)
hold off
% Make Figure of NISE and SESH instantaneous eigenstate population
plot(0:dt:time,P,0:dt:time,PSH,'LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
axis([ 0 time 0 1])
hold on
figure(2)
hold off
% Make Figure of NISE local site population
plot(0:dt:time,PLOC,'LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
axis([ 0 time 0 1])
hold on
toc
end

% Identify auxilary wavefunction state for new timestep
function state=find_present_state(state,C,phin)
% Test previous first
if abs(phin'*C(:,state))>0.5
    return
end
N=length(phin);
% Find state with maximal overlap
ol=0;
state=0;
for i=1:N
    overlap=abs(phin'*C(:,i));
    if overlap>0.5
       state=i;
       return
    elseif overlap>ol
        ol=overlap;
        state=i;
    end
end
end

% Perform surface-hopping (SESH)
function state=hopping(state,oldstate,C,phiA,phin,E,kBT,dt)
N=length(phin);
% Find random number
eta=rand(1);
HP=0;
% Loop over states
for i=1:N
    % Find hopping probability 
    HP0=-2*real((phin'*C(:,i))/(phin'*C(:,state)));
    HP0=HP0*(C(:,state)-phiA)'*C(:,i);
    if HP0<0
        HP0=0;
    end
    HP=HP+HP0;
    % Check if we got a hop
    if HP>eta
        % Now check if hop is valid
        eta2=kBT*abs(randn(1)); % Temperature of non-adiabative cv
        if E(i,i)-E(state,state)<eta2
            % Allowed hop
            state=i;
            return
        else
            % Hop not approved
            return
        end
    end
    % Discard unrealistic hops over 10 kBT
    % if E(i,i)-E(state,state)>10*kBT
       % return
    % end
end
end