%
% TNISE.m
%
% This matlab function calculates the population of the
% instantaneous eigenfunctions using the TNISE method 
% described in this publication
function TNISE()
tic
N=2; % Number of sites (only fully tested for 2)
r=10000; % Number of realizations
dt=10.0; % Time step in fs

icm2ifs=2.99792458e-5; % cm/fs
deltat=dt; % Timestep in fs
factor=icm2ifs*2*pi*deltat;
T=150; % Kelvin 
kBT=T*0.6950389; % cm-1
time=4000; % Length in fs

dw0=70; % Energy Gap (cm-1)
sigma=25; % Fluctuation Magnitude
J=36; % Nearest Neighbor Coupling
tau=50; % Time scale in fs

% Initialize variables
a=exp(-dt/tau);
b=sqrt(1-a*a);
H0=diag(J*ones(N-1),1)+diag(J*ones(N-1),-1);
H0(1,1)=dw0;
w=randn(N,1)*sigma;
PSN=zeros(size(0:dt:time));
PSLOC=zeros(size(0:dt:time));
% Loop over realizations
for rz=1:r
    % Initialize frequencies
    w=randn(N,1)*sigma;
    % Initialize real wavefunction
    phi=zeros(N,1);
    H=H0+diag(w);
    [C,E]=eig(H);
    [C0,E0]=eig(H);
    phi=C(:,N); % Excitation
    Cold=C;
    Eold=E;
    phiLOC=[1 ; 0];
    state=N;
    phiA=phi;
    phiB=phiLOC;
    PSLOC(1)=PSLOC(1)+1;
    PSN(1)=PSN(1)+1;
    step=2;
    % Transfer to instantaneous eigenbasis
    phiA=Cold'*phiA;
    phiB=Cold'*phiB;
    % Loop over times
    for t=dt:dt:time
        % Generate new hamiltonian
        w=w*a+b*randn(N,1)*sigma;
        H=H0+diag(w);
        [C,E]=eig(H);
        % Do scaled NISE step (TNISE)
        U=scaledNISE(C,Cold,E,Eold,factor,N,kBT);
        phiA=U*phiA;
        phiB=U*phiB;
        phiA=renorm(phiA,N);
        phiB=renorm(phiB,N);
        Cold=C;
        Eold=E;
        PSN(step)=PSN(step)+abs(C(:,N)'*C*phiA)^2;
        PSLOC(step)=PSLOC(step)+abs([1 0]*C*phiB)^2;
        step=step+1;
    end
end
PSN=PSN/r;
PSLOC=PSLOC/r;
% Make Figure of TNISE population of instantaneous eigenstate
figure(1)
plot(0:dt:time,PSN,'k--','LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
axis([ 0 time 0 1])
hold on
% Make Figure of TNISE population of localized state
figure(2)
plot(0:dt:time,PSLOC,'k--','LineWidth',2,'MarkerSize',32)
    xlabel('Time in fs','fontsize',18)
    ylabel('Population','fontsize',18)
    set(gca,'fontsize',18)
    set(gca,'linewidth',2)
axis([ 0 time 0 1])
hold on
toc
end

% Subroutine for doing Scaled NISE (TNISE) propagation step
function U=scaledNISE(C,Cold,E,Eold,factor,N,kBT)
CC=C'*Cold;
% Here we will manipulate CC
for ii=1:N
    % Find matching eigenstate
    [ maxC,kk]=max(CC(:,ii).^2);
    % Give warning if the change in the eigenstates is large
    if maxC<0.666
        maxC
        warning('Try smaller timesteps')
    end
    % Do scaling
    CD=0;
    for jj=1:N
        if (jj~=kk)
            DE=E(jj,jj)-Eold(ii,ii);
            if 1>0
                CC(jj,ii)=CC(jj,ii)*exp(-DE/kBT/4);
            end
            CD=CD+CC(jj,ii)*CC(jj,ii)';
        end
    end
    CC(kk,ii)=sqrt(1-CD)*CC(kk,ii)/abs(CC(kk,ii)); % Renormalize
end
U=expm(-i*E*factor)*CC; % Eigenstate propagation
end

% Subroutine for normalization
function phi=renorm(phi,N)
CD=phi'*phi;
phi=phi/sqrt(CD);
end